skhubness.neighbors.classification — scikit-hubness 0.21.0a8 documentation

# Source code for skhubness.neighbors.classification

    # SPDX-License-Identifier: BSD-3-Clause

    """Nearest Neighbor Classification
    adapted from https://github.com/scikit-learn/scikit-learn/blob/0.21.X/sklearn/neighbors/classification.py"""

    # Authors: Jake Vanderplas <vanderplas@astro.washington.edu>
    #          Fabian Pedregosa <fabian.pedregosa@inria.fr>
    #          Alexandre Gramfort <alexandre.gramfort@inria.fr>
    #          Sparseness support by Lars Buitinck
    #          Multi-output support by Arnaud Joly <a.joly@ulg.ac.be>
    #          Hubness support by Roman Feldbauer <roman.feldbauer@univie.ac.at>
    #
    # License: BSD 3 clause (C) INRIA, University of Amsterdam

    import numpy as np
    from scipy import stats
    from sklearn.utils.extmath import weighted_mode

    from sklearn.neighbors.base import SupervisedIntegerMixin
    from sklearn.base import ClassifierMixin
    from sklearn.utils import check_array

    from .base import \
        _check_weights, _get_weights, \
        NeighborsBase, KNeighborsMixin, RadiusNeighborsMixin

    class KNeighborsClassifier(NeighborsBase, KNeighborsMixin,
                               SupervisedIntegerMixin, ClassifierMixin):
        """Classifier implementing the k-nearest neighbors vote.

        Read more in the `scikit-learn User Guide <https://scikit-learn.org/stable/modules/neighbors.html#classification>`_

        Parameters
        ----------
        n_neighbors: int, optional (default = 5)
            Number of neighbors to use by default for :meth:`kneighbors` queries.

        weights: str or callable, optional (default = 'uniform')
            weight function used in prediction.  Possible values:

            - 'uniform': uniform weights.  All points in each neighborhood
              are weighted equally.
            - 'distance': weight points by the inverse of their distance.
              in this case, closer neighbors of a query point will have a
              greater influence than neighbors which are further away.
            - [callable]: a user-defined function which accepts an
              array of distances, and returns an array of the same shape
              containing the weights.

        algorithm : {'auto', 'hnsw', 'lsh', 'falconn_lsh', 'onng', 'rptree', 'ball_tree', 'kd_tree', 'brute'}, optional
            Algorithm used to compute the nearest neighbors:

            - 'hnsw' will use :class:`HNSW`
            - 'lsh' will use :class:`PuffinnLSH`
            - 'falconn_lsh' will use :class:`FalconnLSH`
            - 'onng' will use :class:`NNG`
            - 'rptree' will use :class:`RandomProjectionTree`
            - 'ball_tree' will use :class:`BallTree`
            - 'kd_tree' will use :class:`KDTree`
            - 'brute' will use a brute-force search.
            - 'auto' will attempt to decide the most appropriate exact algorithm
              based on the values passed to :meth:`fit` method. This will not
              select an approximate nearest neighbor algorithm.

            Note: fitting on sparse input will override the setting of
            this parameter, using brute force.

        algorithm_params: dict, optional
            Override default parameters of the NN algorithm.
            For example, with algorithm='lsh' and algorithm_params={n_candidates: 100}
            one hundred approximate neighbors are retrieved with LSH.
            If parameter hubness is set, the candidate neighbors are further reordered
            with hubness reduction.
            Finally, n_neighbors objects are used from the (optionally reordered) candidates.

        hubness: {'mutual_proximity', 'local_scaling', 'dis_sim_local', None}, optional
            Hubness reduction algorithm

            - 'mutual_proximity' or 'mp' will use :class:`MutualProximity`
            - 'local_scaling' or 'ls' will use :class:`LocalScaling`
            - 'dis_sim_local' or 'dsl' will use :class:`DisSimLocal`

            If None, no hubness reduction will be performed (=vanilla kNN).

        hubness_params: dict, optional
            Override default parameters of the selected hubness reduction algorithm.
            For example, with hubness='mp' and hubness_params={'method': 'normal'}
            a mutual proximity variant is used, which models distance distributions
            with independent Gaussians.

        leaf_size: int, optional (default = 30)
            Leaf size passed to BallTree or KDTree.  This can affect the
            speed of the construction and query, as well as the memory
            required to store the tree.  The optimal value depends on the
            nature of the problem.

        p: integer, optional (default = 2)
            Power parameter for the Minkowski metric. When p = 1, this is
            equivalent to using manhattan_distance (l1), and euclidean_distance
            (l2) for p = 2. For arbitrary p, minkowski_distance (l_p) is used.

        metric: string or callable, default 'minkowski'
            the distance metric to use for the tree.  The default metric is
            minkowski, and with p=2 is equivalent to the standard Euclidean
            metric. See the documentation of the DistanceMetric class for a
            list of available metrics.

        metric_params: dict, optional (default = None)
            Additional keyword arguments for the metric function.

        n_jobs: int or None, optional (default=None)
            The number of parallel jobs to run for neighbors search.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
            ``-1`` means using all processors.
            See `Glossary <https://scikit-learn.org/stable/glossary.html#term-n-jobs>`_ for more details.
            Doesn't affect :meth:`fit` method.

        Examples
        --------
        >>> X = [[0], [1], [2], [3]]
        >>> y = [0, 0, 1, 1]
        >>> from skhubness.neighbors import KNeighborsClassifier
        >>> neigh = KNeighborsClassifier(n_neighbors=3)
        >>> neigh.fit(X, y) # doctest: +ELLIPSIS
        KNeighborsClassifier(...)
        >>> print(neigh.predict([[1.1]]))
        [0]
        >>> print(neigh.predict_proba([[0.9]]))
        [[0.66666667 0.33333333]]

        See also
        --------
        RadiusNeighborsClassifier
        KNeighborsRegressor
        RadiusNeighborsRegressor
        NearestNeighbors

        Notes
        -----
        See `Nearest Neighbors <https://scikit-learn.org/stable/modules/neighbors.html#neighbors>`_
        in the scikit-learn online documentation for a discussion
        of the choice of ``algorithm`` and ``leaf_size``.

        .. warning::
           Regarding the Nearest Neighbors algorithms, if it is found that two
           neighbors, neighbor `k+1` and `k`, have identical distances
           but different labels, the results will depend on the ordering of the
           training data.

        https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm
        """

        def __init__(self, n_neighbors: int = 5, weights: str = 'uniform',
                     algorithm: str = 'auto', algorithm_params: dict = None,
                     hubness: str = None, hubness_params: dict = None,
                     leaf_size: int = 30, p=2, metric='minkowski', metric_params=None,
                     n_jobs=None, verbose: int = 0, **kwargs):

            super().__init__(
                n_neighbors=n_neighbors,
                algorithm=algorithm,
                algorithm_params=algorithm_params,
                hubness=hubness,
                hubness_params=hubness_params,
                leaf_size=leaf_size, metric=metric, p=p,
                metric_params=metric_params,
                n_jobs=n_jobs,
                verbose=verbose,
                **kwargs)
            self.weights = _check_weights(weights)

        def predict(self, X):
            """Predict the class labels for the provided data

            Parameters
            ----------
            X: array-like, shape (n_query, n_features), \
                    or (n_query, n_indexed) if metric == 'precomputed'
                Test samples.

            Returns
            -------
            y: array of shape [n_samples] or [n_samples, n_outputs]
                Class labels for each data sample.
            """
            X = check_array(X, accept_sparse='csr')

            neigh_dist, neigh_ind = self.kneighbors(X)
            classes_ = self.classes_
            _y = self._y
            if not self.outputs_2d_:
                _y = self._y.reshape((-1, 1))
                classes_ = [self.classes_]

            n_outputs = len(classes_)
            n_samples = X.shape[0]
            weights = _get_weights(neigh_dist, self.weights)

            y_pred = np.empty((n_samples, n_outputs), dtype=classes_[0].dtype)
            for k, classes_k in enumerate(classes_):
                if weights is None:
                    mode, _ = stats.mode(_y[neigh_ind, k], axis=1)
                else:
                    mode, _ = weighted_mode(_y[neigh_ind, k], weights, axis=1)

                mode = np.asarray(mode.ravel(), dtype=np.intp)
                y_pred[:, k] = classes_k.take(mode)

            if not self.outputs_2d_:
                y_pred = y_pred.ravel()

            return y_pred

        def predict_proba(self, X):
            """Return probability estimates for the test data X.

            Parameters
            ----------
            X: array-like, shape (n_query, n_features), \
                    or (n_query, n_indexed) if metric == 'precomputed'
                Test samples.

            Returns
            -------
            p: array of shape = [n_samples, n_classes], or a list of n_outputs
                of such arrays if n_outputs > 1.
                The class probabilities of the input samples. Classes are ordered
                by lexicographic order.
            """
            X = check_array(X, accept_sparse='csr')

            neigh_dist, neigh_ind = self.kneighbors(X)

            classes_ = self.classes_
            _y = self._y
            if not self.outputs_2d_:
                _y = self._y.reshape((-1, 1))
                classes_ = [self.classes_]

            n_samples = X.shape[0]

            weights = _get_weights(neigh_dist, self.weights)
            if weights is None:
                weights = np.ones_like(neigh_ind)

            all_rows = np.arange(X.shape[0])
            probabilities = []
            for k, classes_k in enumerate(classes_):
                pred_labels = _y[:, k][neigh_ind]
                proba_k = np.zeros((n_samples, classes_k.size))

                # a simple ':' index doesn't work right
                for i, idx in enumerate(pred_labels.T):  # loop is O(n_neighbors)
                    proba_k[all_rows, idx] += weights[:, i]

                # normalize 'votes' into real [0,1] probabilities
                normalizer = proba_k.sum(axis=1)[:, np.newaxis]
                normalizer[normalizer == 0.0] = 1.0
                proba_k /= normalizer

                probabilities.append(proba_k)

            if not self.outputs_2d_:
                probabilities = probabilities[0]

            return probabilities

    class RadiusNeighborsClassifier(NeighborsBase, RadiusNeighborsMixin,
                                    SupervisedIntegerMixin, ClassifierMixin):
        """Classifier implementing a vote among neighbors within a given radius

        Read more in the `scikit-learn User Guide
        <https://scikit-learn.org/stable/modules/neighbors.html#classification>`_

        Parameters
        ----------
        radius: float, optional (default = 1.0)
            Range of parameter space to use by default for :meth:`radius_neighbors`
            queries.

        weights: str or callable
            weight function used in prediction.  Possible values:

            - 'uniform': uniform weights.  All points in each neighborhood
              are weighted equally.
            - 'distance': weight points by the inverse of their distance.
              in this case, closer neighbors of a query point will have a
              greater influence than neighbors which are further away.
            - [callable]: a user-defined function which accepts an
              array of distances, and returns an array of the same shape
              containing the weights.

            Uniform weights are used by default.

        algorithm: {'auto', 'falconn_lsh', 'ball_tree', 'kd_tree', 'brute'}, optional
            Algorithm used to compute the nearest neighbors:

            - 'falconn_lsh' will use :class:`FalconnLSH`
            - 'ball_tree' will use :class:`BallTree`
            - 'kd_tree' will use :class:`KDTree`
            - 'brute' will use a brute-force search.
            - 'auto' will attempt to decide the most appropriate algorithm
              based on the values passed to :meth:`fit` method.

            Note: fitting on sparse input will override the setting of
            this parameter, using brute force.

        algorithm_params: dict, optional
            Override default parameters of the NN algorithm.
            For example, with algorithm='lsh' and algorithm_params={n_candidates: 100}
            one hundred approximate neighbors are retrieved with LSH.
            If parameter hubness is set, the candidate neighbors are further reordered
            with hubness reduction.
            Finally, n_neighbors objects are used from the (optionally reordered) candidates.

        hubness: {'mutual_proximity', 'local_scaling', 'dis_sim_local', None}, optional
            Hubness reduction algorithm

            - 'mutual_proximity' or 'mp' will use :class:`MutualProximity`
            - 'local_scaling' or 'ls' will use :class:`LocalScaling`
            - 'dis_sim_local' or 'dsl' will use :class:`DisSimLocal`

            If None, no hubness reduction will be performed (=vanilla kNN).

        hubness_params: dict, optional
            Override default parameters of the selected hubness reduction algorithm.
            For example, with hubness='mp' and hubness_params={'method': 'normal'}
            a mutual proximity variant is used, which models distance distributions
            with independent Gaussians.

        leaf_size: int, optional (default = 30)
            Leaf size passed to BallTree or KDTree.  This can affect the
            speed of the construction and query, as well as the memory
            required to store the tree.  The optimal value depends on the
            nature of the problem.

        p: integer, optional (default = 2)
            Power parameter for the Minkowski metric. When p = 1, this is
            equivalent to using manhattan_distance (l1), and euclidean_distance
            (l2) for p = 2. For arbitrary p, minkowski_distance (l_p) is used.

        metric: string or callable, default 'minkowski'
            the distance metric to use for the tree.  The default metric is
            minkowski, and with p=2 is equivalent to the standard Euclidean
            metric. See the documentation of the DistanceMetric class for a
            list of available metrics.

        outlier_label: int, optional (default = None)
            Label, which is given for outlier samples (samples with no
            neighbors on given radius).
            If set to None, ValueError is raised, when outlier is detected.

        metric_params: dict, optional (default = None)
            Additional keyword arguments for the metric function.

        n_jobs: int or None, optional (default=None)
            The number of parallel jobs to run for neighbors search.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
            ``-1`` means using all processors.
            See `Glossary <https://scikit-learn.org/stable/glossary.html#term-n-jobs>`_
            for more details.

        Examples
        --------
        >>> X = [[0], [1], [2], [3]]
        >>> y = [0, 0, 1, 1]
        >>> from skhubness.neighbors import RadiusNeighborsClassifier
        >>> neigh = RadiusNeighborsClassifier(radius=1.0)
        >>> neigh.fit(X, y) # doctest: +ELLIPSIS
        RadiusNeighborsClassifier(...)
        >>> print(neigh.predict([[1.5]]))
        [0]

        See also
        --------
        KNeighborsClassifier
        RadiusNeighborsRegressor
        KNeighborsRegressor
        NearestNeighbors

        Notes
        -----
        See `Nearest Neighbors <https://scikit-learn.org/stable/modules/neighbors.html#neighbors>`_
        in the scikit-learn online documentation for a discussion
        of the choice of ``algorithm`` and ``leaf_size``.

        https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm
        """

        def __init__(self, radius=1.0, weights='uniform',
                     algorithm: str = 'auto', algorithm_params: dict = None,
                     hubness: str = None, hubness_params: dict = None,
                     leaf_size=30, p=2, metric='minkowski',
                     outlier_label=None, metric_params=None, n_jobs=None,
                     **kwargs):
            super().__init__(
                radius=radius,
                algorithm=algorithm,
                algorithm_params=algorithm_params,
                hubness=hubness,
                hubness_params=hubness_params,
                leaf_size=leaf_size,
                metric=metric, p=p, metric_params=metric_params,
                n_jobs=n_jobs, **kwargs)
            self.weights = _check_weights(weights)
            self.outlier_label = outlier_label

        def predict(self, X):
            """Predict the class labels for the provided data

            Parameters
            ----------
            X: array-like, shape (n_query, n_features), \
                    or (n_query, n_indexed) if metric == 'precomputed'
                Test samples.

            Returns
            -------
            y: array of shape [n_samples] or [n_samples, n_outputs]
                Class labels for each data sample.
            """
            X = check_array(X, accept_sparse='csr')
            n_samples = X.shape[0]

            neigh_dist, neigh_ind = self.radius_neighbors(X)
            inliers = [i for i, nind in enumerate(neigh_ind) if len(nind) != 0]
            outliers = [i for i, nind in enumerate(neigh_ind) if len(nind) == 0]

            classes_ = self.classes_
            _y = self._y
            if not self.outputs_2d_:
                _y = self._y.reshape((-1, 1))
                classes_ = [self.classes_]
            n_outputs = len(classes_)

            if self.outlier_label is not None:
                neigh_dist[outliers] = 1e-6
            elif outliers:
                raise ValueError('No neighbors found for test samples %r, '
                                 'you can try using larger radius, '
                                 'give a label for outliers, '
                                 'or consider removing them from your dataset.'
                                 % outliers)

            weights = _get_weights(neigh_dist, self.weights)

            y_pred = np.empty((n_samples, n_outputs), dtype=classes_[0].dtype)
            for k, classes_k in enumerate(classes_):
                pred_labels = np.zeros(len(neigh_ind), dtype=object)
                pred_labels[:] = [_y[ind, k] for ind in neigh_ind]
                if weights is None:
                    mode = np.array([stats.mode(pl)[0]
                                     for pl in pred_labels[inliers]], dtype=np.int)
                else:
                    mode = np.array(
                        [weighted_mode(pl, w)[0]
                         for (pl, w) in zip(pred_labels[inliers], weights[inliers])
                         ], dtype=np.int)

                mode = mode.ravel()

                y_pred[inliers, k] = classes_k.take(mode)

            if outliers:
                y_pred[outliers, :] = self.outlier_label

            if not self.outputs_2d_:
                y_pred = y_pred.ravel()

            return y_pred